package	jp.co.sds_corp.util.daemon.core;

import java.io.Serializable;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Timer;
import java.util.TimerTask;
import java.util.Vector;

import jp.co.sds_corp.util.common.CmnDate;
import jp.co.sds_corp.util.common.CmnLog;
import jp.co.sds_corp.util.config.IConfiguration;

/**
* Cronデーモンの実装です。
* 指定された時間ごとに、サービスを呼び出します。
<P>
XMLは以下のような形式です。
<PRE>
&lt;Daemon name="Daily" class="jp.co.sds_corp.util.daemon.core.CronDaemon"&gt;
	&lt;Schedule&gt;
		&lt;time&gt;03:00:00&lt;/time&gt;
	&lt;/Schedule&gt;
	&lt;Service name="Daily" class="daily.DailyUpdate"&gt;
		&lt;dbkey&gt;test&lt;/dbkey&gt;
		...
	&lt;/Service&gt;
&lt;/Daemon&gt;
</PRE>
&lt;Schedule&gt;要素の記述方法については、{@link Schedule}を参照してください。
*/
public class CronDaemon extends AbstractDaemon
		implements CronDaemonMBean
{
	private transient Timer timer = null;
	private Vector<Schedule> vecSch = new Vector<Schedule>();
	private Map<Schedule,ServiceTask> mapTask = new HashMap<Schedule,ServiceTask>();

	/**
	* 設定情報でこのデーモンを初期化します。
	*/
	public void init( IConfiguration config )
		throws Exception
	{
		super.init( config );
		Iterator<IConfiguration> itr = config.getChilds( "Schedule" );
		while ( itr.hasNext()) {
			IConfiguration sc = (IConfiguration )itr.next();
			Schedule sch = new Schedule();
			sch.init( sc );
			vecSch.add( sch );
		}
	}

	/**
	* Cronデーモンを開始します。
	*/
	public void start() {
		if ( getStatus() == Status.ERROR ) {
			return;
		}
		timer = new Timer();
		for ( int i = 0; i < vecSch.size(); i++ ) {
			Schedule sch = (Schedule )vecSch.get( i );
			start( sch );
		}
		setStatus( Status.ALIVE );
	}

	/**
	* 指定されたスケジュールで、Cronデーモンを開始します。
	* @param	sch		スケジュール
	*/
	protected void start( Schedule sch ) {
		ServiceTask task = new ServiceTask( sch );
		mapTask.put( sch, task );
		switch ( sch.getCycle()) {
		case WEEK:
		case DAY:
		case HOUR:
		case MIN:
			timer.scheduleAtFixedRate( task, sch.getNextTime().getTime(),
				sch.getInterval() );
			break;
		case MONTH:
		case ONCE:
			if ( sch.getNextTime().compareTo( new CmnDate()) > 0 ) {
				timer.schedule( task, sch.getNextTime().getTime() );
			}
			break;
		}
	}

	/**
	* スケジュールを追加します。
	* @param	sch		スケジュール
	*/
	public void addSchedule( Schedule sch ) {
		vecSch.add( sch );
		start( sch );
	}

	/**
	* スケジュールを追加します。
	* @param	idx		スケジュールのインデックス
	*/
	public Schedule removeSchedule( int idx ) {
		Schedule sch = (Schedule )vecSch.remove( idx );
		ServiceTask task = (ServiceTask )mapTask.remove( sch );
		if ( task != null ) {
			task.cancel();
		}
		return sch;
	}

	/**
	* スケジュールの一覧を返す。
	* @return	スケジュールの一覧
	*/
	public List<Schedule> listSchedule() {
		return vecSch;
	}

	/**
	* スケジュールと関係なく、直ちに処理を実行します。
	*/
	public void startImmediate() throws Exception {
		Schedule sch = new Schedule();
		sch.setTime( "now" );
		ServiceTask task = new ServiceTask( sch );
		timer.schedule( task, sch.getNextTime().getTime() );
	}

	/**
	* Cronデーモンに割り込みます。
	*/
	public void terminate() {
		CmnLog.debug( this, "terminate:" + getName() + ":" + getClass().getName( ) );
		for ( ServiceTask task : mapTask.values() ) {
			task.cancel();
		}
		mapTask.clear();
		if ( timer != null ) {
			timer.cancel();
		}
		getService().stop();
		setStatus( Status.DEAD );
	}

	/**
	* このデーモンの全ての Thread の終了を待ちます。
	*/
	public void joinAll() {
	}

	/**
	* 設定されている内容をXML形式で出力します。
	* @param	config	設定情報
	*/
	public void saveConfig( IConfiguration config ) {
		super.saveConfig( config );
		for ( int i = 0; i < vecSch.size(); i++ ) {
			Schedule sch = (Schedule)vecSch.get( i );
			IConfiguration c = config.addChild( "Schedule" );
			sch.saveConfig( c );
		}
	}

	/**
	* Cronデーモンの文字列表現を返します。
	* @return	Cronデーモンの文字列表現
	*/
	public String toString() {
		String ret = super.toString();
		for ( int i = 0; i < vecSch.size(); i++ ) {
			ret += "\n\t" + ((Schedule)vecSch.get( i )).toString();
		}
		return ret;
	}

	/**
	* Cronの実際の処理を行うTimerTask。
	*/
	private class ServiceTask extends TimerTask implements Serializable
	{
		Schedule sch;
		public ServiceTask( Schedule sch ) {
			this.sch = sch;
		}

		public void run() {
			if ( isDebug()) {
				CmnLog.debug( this, "start CRON service:" + getService() );
			}
			setStatus( Status.RUN );
			try {
				Thread.currentThread().setName( CronDaemon.this.getName()
						+ ":" + CronDaemon.this.getService().getName() );
				getService().start();
				switch ( sch.getCycle()) {
				case MONTH:
					timer.schedule( this, sch.getNextTime().getTime() );
					break;
				}
			} catch ( Throwable t ) {
				CmnLog.error( this, "Service error", t );
				setStatus( Status.ERROR, t.getMessage() );
			}
			setStatus( Status.ALIVE );
			if ( isDebug()) {
				CmnLog.debug( this, "end CRON service:" + getService() );
			}
		}
	}
}

