package	jp.co.sds_corp.util.daemon.core;

import jp.co.sds_corp.util.common.CmnLog;
import jp.co.sds_corp.util.config.IConfiguration;

/**
* デーモンクラス。
* サービスを別Threadを使用して起動します。
* このクラスを利用したサービスは、start()からreturnすると
* 異常終了と見なされます。
<P>
XMLは以下の形式です。
<PRE>
&lt;Daemon name="DAEMON" class="jp.co.sds_corp.util.daemon.core.Daemon"&gt;
	&lt;return&gt;true&lt;/return&gt;
	&lt;Service name="service" class="daemon.ServiceTest"&gt;
		...
	&lt;/Service&gt;
&lt;/Daemon&gt;
</PRE>
*/
public class Daemon extends AbstractDaemon
		implements DaemonMBean
{
	protected transient Thread	thr;
	protected boolean	bRet = false;

	/**
	* デーモンを初期化します。
	*/
	public void init( IConfiguration config ) throws Exception {
		super.init( config );
		String err = config.get( "return", null );
		if ( err != null && err.equals( "true" )) {
			bRet = true;
		}
	}

	/**
	* デーモンを開始します。
	*/
	public void start() {
		if ( getStatus() == Status.ERROR ) {
			return;
		}
		thr = new ServiceThread();
		thr.start();
		setStatus( Status.ALIVE );
		getService().setInterrupt( false );
	}

	/**
	* デーモンに割り込みます。
	*/
	public void terminate() {
		getService().setInterrupt( true );
		thr.interrupt();
		getService().stop();
		setStatus( Status.DEAD, "Terminated." );
	}

	/**
	* 全てのThreadの終了を待ちます。
	*/
	public void joinAll() {
		try {
			thr.join();
		} catch ( Exception ex ) {
		}
	}

	/**
	* 設定されている内容をXML形式で出力します。
	* @param	config	設定情報
	*/
	public void saveConfig( IConfiguration config ) {
		super.saveConfig( config );
		if ( bRet ) {
			config.addChild( "return", "true" );
		}
	}

	/**
	* 処理を起動するThreadを管理します。
	*/
	private class ServiceThread extends Thread
	{
		public ServiceThread() {
		}

		public void run() {
			CmnLog.debug( this, "start DAEMON service:" + getService() );
			setStatus( Status.RUN );
			Thread.currentThread().setName( Daemon.this.getName()
					+ ":" + Daemon.this.getService().getName() );
			try {
				getService().start();
			} catch ( Throwable t ) {
				String msg = "Service error "
					+ "[" + Daemon.this.getName()
					+ ":" + Daemon.this.getService().getName() + "]";
				CmnLog.error( this, msg, t );
				Daemon.this.setStatus( Status.DEAD, t.getMessage() );
			}
			if ( !bRet ) {
				String msg = "Service " + getService() + " stopped.";
				if ( getService().isInterrupted()) {
					CmnLog.warn( this, msg );
				} else {
					CmnLog.error( this, msg );
				}
				Daemon.this.setStatus( Status.DEAD );
			}
		}
	}
}

