package	jp.co.sds_corp.util.daemon.core;

import java.io.Serializable;
import java.text.DateFormatSymbols;

import jp.co.sds_corp.util.common.CmnDate;
import jp.co.sds_corp.util.config.IConfigSavable;
import jp.co.sds_corp.util.config.IConfigurable;
import jp.co.sds_corp.util.config.IConfiguration;

/**
* Cronデーモンで使用されるスケジュールを定義します。
<P>
XMLは以下のような形式を選択できます。
<BR>
一回のみ起動する場合は、以下の２通りの記述が可能です。
<PRE>
&lt;Schedule&gt;
	&lt;time&gt;2003/12/08 19:26:00&lt;/time&gt;
&lt;/Schedule&gt;
&lt;Schedule&gt;
	&lt;year&gt;2003&lt;/year&gt;
	&lt;month&gt;12&lt;/month&gt;
	&lt;date&gt;31&lt;/date&gt;
	&lt;dayofweek&gt;日&lt;/dayofweek&gt;
	&lt;hour&gt;12&lt;/hour&gt;
	&lt;min&gt;31&lt;/min&gt;
	&lt;sec&gt;31&lt;/sec&gt;
&lt;/Schedule&gt;
</PRE>

&lt;time&gt;要素は、時間のみ(HH:mm:ss)の指定も可能です。
曜日も指定すると週次で起動されます。
<PRE>
&lt;Schedule&gt;
	&lt;dayofweek&gt;金&lt;/dayofweek&gt;
	&lt;time&gt;18:00:00&lt;/time&gt;
&lt;/Schedule&gt;
</PRE>

一日一回の起動です。
<PRE>
&lt;Schedule&gt;
	&lt;hour&gt;03&lt;/hour&gt;
	&lt;min&gt;00&lt;/min&gt;
	&lt;sec&gt;00&lt;/sec&gt;
&lt;/Schedule&gt;
</PRE>

より上位のパラメータは省略することが可能です。
以下の例では毎分３１秒に起動されます。
<PRE>
&lt;Schedule&gt;
	&lt;sec&gt;31&lt;/sec&gt;
&lt;/Schedule&gt;
</PRE>
*/
public class Schedule implements Serializable, IConfigurable, IConfigSavable
{
	/** 起動周期 */
	public enum Cycle {
		/** 起動周期：不明 */
		UNKNOWN,
		/** 起動周期：一回きり */
		ONCE,
		/** 起動周期：年 */
		YEAR,
		/** 起動周期：月 */
		MONTH,
		/** 起動周期：日 */
		DAY,
		/** 起動周期：週 */
		WEEK,
		/** 起動周期：時 */
		HOUR,
		/** 起動周期：分 */
		MIN,
		/** 起動周期：秒 */
		SEC;
	}
	/** 起動周期 */
	public Cycle cycle = Cycle.UNKNOWN;

	private CmnDate	time;
	private CmnDate	next;
	private int	year = -1;
	private int	month = -1;
	private int	date = -1;
	private int	dayofweek = -1;
	private int	hour = -1;
	private int	min = -1;
	private int	sec = -1;
	String timestr = null;

	/**
	* コンストラクタ
	*/
	public Schedule() {
	}

	/**
	* コンストラクタ
	* @param	cycle	起動周期
	*/
	public Schedule( Cycle cycle ) {
		this.cycle = cycle;
	}

	/**
	* 設定情報からスケジュールを生成します。
	*/
	public void init( IConfiguration conf ) throws Exception {
		String var;
		var = conf.get( "year", null );
		if ( var != null ) {
			year = Integer.valueOf( var ).intValue();
			if ( cycle == Cycle.UNKNOWN ) {
				cycle = Cycle.ONCE;
			}
		}
		var = conf.get( "month", null );
		if ( var != null ) {
			month = Integer.valueOf( var ).intValue();
			if ( cycle == Cycle.UNKNOWN ) {
				cycle = Cycle.YEAR;
			}
		}
		var = conf.get( "date", null );
		if ( var != null ) {
			date = Integer.valueOf( var ).intValue();
			if ( cycle == Cycle.UNKNOWN ) {
				cycle = Cycle.MONTH;
			}
		}
		var = conf.get( "dayofweek", null );
		if ( var != null ) {
			DateFormatSymbols dfs = new DateFormatSymbols();
			String[] Weekdays = dfs.getShortWeekdays();
			for ( int i = 0; i < Weekdays.length; i++ ) {
				if ( var.equals( Weekdays[i] )) {
					dayofweek = i;
					if ( cycle == Cycle.UNKNOWN ) {
						cycle = Cycle.WEEK;
					}
					break;
				}
			}
		}
		var = conf.get( "time", null );
		if ( var != null ) {
			setTime( var );
		}
		var = conf.get( "hour", null );
		if ( var != null ) {
			hour = Integer.valueOf( var ).intValue();
			if ( cycle == Cycle.UNKNOWN ) {
				cycle = Cycle.DAY;
			}
		}
		var = conf.get( "min", null );
		if ( var != null ) {
			min = Integer.valueOf( var ).intValue();
			if ( cycle == Cycle.UNKNOWN ) {
				cycle = Cycle.HOUR;
			}
		}
		var = conf.get( "sec", null );
		if ( var != null ) {
			sec = Integer.valueOf( var ).intValue();
			if ( cycle == Cycle.UNKNOWN ) {
				cycle = Cycle.MIN;
			}
		}

		switch ( cycle ) {
		case ONCE:
			if ( time == null && year < 0 ) {
				throw( new IllegalArgumentException(
							"秒が指定されていません:" + toString()));
			}
		case YEAR:
			if ( month < 0 ) {
				throw( new IllegalArgumentException(
							"月が指定されていません:" + toString()));
			}
		case MONTH:
			if ( date < 0 ) {
				throw( new IllegalArgumentException(
							"日が指定されていません:" + toString()));
			}
		case WEEK:
		case DAY:
			if ( hour < 0 ) {
				throw( new IllegalArgumentException(
							"時が指定されていません:" + toString()));
			}
			if ( hour >= 24 ) {
				throw( new IllegalArgumentException(
							"時が不正です:" + toString()));
			}
		case HOUR:
			if ( min < 0 ) {
				throw( new IllegalArgumentException(
							"分が指定されていません:" + toString()));
			}
			if ( min >= 60 ) {
				throw( new IllegalArgumentException(
							"分が不正です:" + toString()));
			}
		case MIN:
			if ( sec < 0 ) {
				throw( new IllegalArgumentException(
							"秒が指定されていません:" + toString()));
			}
			if ( sec >= 60 ) {
				throw( new IllegalArgumentException(
							"秒が不正です:" + toString()));
			}
			break;
		}
	}

	/**
	* 指定時刻(time)を設定します。
	* @param	var		設定するデータ
	*/
	public void setTime( String var ) throws Exception {
		this.timestr = var;
		if ( var.equals( "now" )) {
			CmnDate tmp = new CmnDate();
			tmp.add( CmnDate.SECOND, 3 );
			setTime( tmp );
		} else {
			try {
				setTime( new CmnDate( var, "yyyy/MM/dd HH:mm:ss" ));
			} catch ( Exception de1 ) {
			}
		}
		if ( time != null ) {
			cycle = Cycle.ONCE;
		} else {
			time = new CmnDate( var, "HH:mm:ss" );
			if ( cycle == Cycle.UNKNOWN ) {
				cycle = Cycle.DAY;
			}
			hour = time.get( CmnDate.HOUR_OF_DAY );
			min = time.get( CmnDate.MINUTE );
			sec = time.get( CmnDate.SECOND );
		}
	}
	/**
	* 指定時刻(time)を設定します。
	* @param	var		設定するデータ
	*/
	public void setTime( CmnDate var ) {
		this.time = var ;
		year = var.get( CmnDate.YEAR );
		month = var.get( CmnDate.MONTH ) + 1;
		date = var.get( CmnDate.DATE );
		hour = var.get( CmnDate.HOUR_OF_DAY );
		min = var.get( CmnDate.MINUTE );
		sec = var.get( CmnDate.SECOND );
	}
	/**
	* 指定時刻(time)を取得します。
	* @return	timeの内容
	*/
	public CmnDate getTime() {
		return time;
	}

	/**
	* 年(year)を設定します。
	* @param	var		設定するデータ
	*/
	public void setYear( int var ) {
		this.year = var ;
	}
	/**
	* 年(year)を取得します。
	* @return	yearの内容
	*/
	public int getYear() {
		return year;
	}

	/**
	* 月(month)を設定します。
	* @param	var		設定するデータ
	*/
	public void setMonth( int var ) {
		this.month = var ;
	}
	/**
	* 月(month)を取得します。
	* @return	monthの内容
	*/
	public int getMonth() {
		return month;
	}

	/**
	* 日(date)を設定します。
	* @param	var		設定するデータ
	*/
	public void setDate( int var ) {
		this.date = var ;
	}
	/**
	* 日(date)を取得します。
	* @return	dateの内容
	*/
	public int getDate() {
		return date;
	}

	/**
	* 曜日(dayofweek)を設定します。
	* @param	var		設定するデータ
	*/
	public void setDayofweek( int var ) {
		this.dayofweek = var ;
	}
	/**
	* 曜日(dayofweek)を取得します。
	* @return	dayofweekの内容
	*/
	public int getDayofweek() {
		return dayofweek;
	}

	/**
	* 時(hour)を設定します。
	* @param	var		設定するデータ
	*/
	public void setHour( int var ) {
		this.hour = var ;
	}
	/**
	* 時(hour)を取得します。
	* @return	hourの内容
	*/
	public int getHour() {
		return hour;
	}

	/**
	* 分(min)を設定します。
	* @param	var		設定するデータ
	*/
	public void setMin( int var ) {
		this.min = var ;
	}
	/**
	* 分(min)を取得します。
	* @return	minの内容
	*/
	public int getMin() {
		return min;
	}

	/**
	* 秒(sec)を設定します。
	* @param	var		設定するデータ
	*/
	public void setSec( int var ) {
		this.sec = var ;
	}
	/**
	* 秒(sec)を取得します。
	* @return	secの内容
	*/
	public int getSec() {
		return sec;
	}

	/**
	* 起動周期(cycle)を取得します。
	* @see		Cycle#UNKNOWN
	* @see		Cycle#ONCE
	* @see		Cycle#YEAR
	* @see		Cycle#MONTH
	* @see		Cycle#DAY
	* @see		Cycle#WEEK
	* @see		Cycle#HOUR
	* @see		Cycle#MIN
	* @see		Cycle#SEC
	* @return	secの内容
	*/
	public Cycle getCycle() {
		return cycle;
	}

	/**
	* サービスを起動する間隔を取得します。
	* @return	サービスを起動する間隔
	*/
	public long getInterval() {
		long lPeriod = -1;
		switch ( cycle ) {
		case ONCE:
		case MONTH:
			break;
		case WEEK:
			lPeriod = 7 * 24 * 60 * 60 * 1000;
			break;
		case DAY:
			lPeriod = 24 * 60 * 60 * 1000;
			break;
		case HOUR:
			lPeriod = 60 * 60 * 1000;
			break;
		case MIN:
			lPeriod = 60 * 1000;
			break;
		}
		return lPeriod;
	}

	/**
	* 次回にサービスを起動する時刻を取得します。
	* @return	次回にサービスを起動する時刻
	*/
	public CmnDate getNextTime() {
		CmnDate now = new CmnDate();
		CmnDate next = new CmnDate();
		switch ( cycle ) {
		case ONCE:
			next.set( CmnDate.YEAR, year );
		case YEAR:
			next.set( CmnDate.MONTH, month - 1 );
		case MONTH:
			next.set( CmnDate.DATE, date );
		case WEEK:
		case DAY:
			next.set( CmnDate.HOUR_OF_DAY, hour );
		case HOUR:
			next.set( CmnDate.MINUTE, min );
		case MIN:
			next.set( CmnDate.SECOND, sec );
		}
		if ( now.compareTo( next ) > 0 ) {
			switch ( cycle ) {
			case WEEK:
				for ( int i = 0; i < 7; i++ ) {
					if ( next.get( CmnDate.DAY_OF_WEEK ) == dayofweek ) {
						break;
					}
					next.add( CmnDate.DATE, 1 );
				}
				break;
			case YEAR: next.add( CmnDate.YEAR, 1 ); break;
			case MONTH: next.add( CmnDate.MONTH, 1 ); break;
			case DAY: next.add( CmnDate.DATE, 1 ); break;
			case HOUR: next.add( CmnDate.HOUR_OF_DAY, 1 ); break;
			case MIN: next.add( CmnDate.MINUTE, 1 ); break;
			}
		}
		return next;
	}

	/**
	* Schedule同士を比較します。
	* @param	var		比較対象のSchedule
	*/
	public int compareTo( Schedule var ) {
		if ( getCycle() != var.getCycle() ) {
			throw( new IllegalArgumentException( "cycle not match" ));
		}
		switch ( cycle ) {
		case ONCE:
			return getTime().compareTo( var.getTime() );
		case YEAR:
			if ( getMonth() != var.getMonth() ) {
				return getMonth() - var.getMonth();
			}
		case MONTH:
			if ( getDate() != var.getDate() ) {
				return getDate() - var.getDate();
			}
		case WEEK:
			if ( getDayofweek() != var.getDayofweek() ) {
				return getDayofweek() - var.getDayofweek();
			}
		case DAY:
			if ( getHour() != var.getHour() ) {
				return getHour() - var.getHour();
			}
		case HOUR:
			if ( getMin() != var.getMin() ) {
				return getMin() - var.getMin();
			}
		case MIN:
			if ( getSec() != var.getSec() ) {
				return getSec() - var.getSec();
			}
		}
		return 0;
	}

	/**
	* Scheduleと日付を比較します。
	* @param	var		比較対象のCmnDate
	*/
	public int compareTo( CmnDate var ) {
		switch ( cycle ) {
		case ONCE:
			return getTime().compareTo( var );
		case YEAR:
			if ( getMonth() != var.get( CmnDate.MONTH ) ) {
				return getMonth() - var.get( CmnDate.MONTH );
			}
		case MONTH:
			if ( getDate() != var.get( CmnDate.DATE ) ) {
				return getDate() - var.get( CmnDate.DATE );
			}
		case WEEK:
			if ( getDayofweek() != -1 &&
				getDayofweek() != var.get( CmnDate.DAY_OF_WEEK ) ) {
				return getDayofweek() - var.get( CmnDate.DAY_OF_WEEK );
			}
		case DAY:
			if ( getHour() != var.get( CmnDate.HOUR_OF_DAY ) ) {
				return getHour() - var.get( CmnDate.HOUR_OF_DAY );
			}
		case HOUR:
			if ( getMin() != var.get( CmnDate.MINUTE ) ) {
				return getMin() - var.get( CmnDate.MINUTE );
			}
		case MIN:
			if ( getSec() != var.get( CmnDate.SECOND ) ) {
				return getSec() - var.get( CmnDate.SECOND );
			}
		}
		return 0;
	}

	/**
	* スケジュールの文字列表現を返します。
	* @return	スケジュールの文字列表現
	*/
	public String toString() {
		String ret = getClass().getName() + "[" + "cicle=" + cycle;
		if ( time == null ) {
			ret += " time=null,";
		} else {
			ret += " time=" + time.format( "yyyy/MM/dd HH:mm:ss" ) + ",";
		}
		ret += "year=" + year + ","
			+ "month=" + month + ","
			+ "date=" + date + ","
			+ "dayofweek=" + dayofweek + ","
			+ "hour=" + hour + ","
			+ "min=" + min + ","
			+ "sec=" + sec + ":"
			+ "next=" + getNextTime().format( "yyyy/MM/dd HH:mm:ss" ) + ","
			+ "interval=" + getInterval() + "]";
		return ret;
	}

	/**
	* 設定されている内容をXML形式で出力します。
	* @param	config	設定情報
	*/
	public void saveConfig( IConfiguration config ) {
		if ( dayofweek >= 0 ) {
			DateFormatSymbols dfs = new DateFormatSymbols();
			String[] Weekdays = dfs.getShortWeekdays();
			config.addChild( "dayofweek", Weekdays[dayofweek] );
		}
		if ( timestr != null ) {
			config.addChild( "time", timestr );
		} else {
			if ( year >= 0 ) {
				config.addChild( "year", String.valueOf( year ));
			}
			if ( month >= 0 ) {
				config.addChild( "month", String.valueOf( month ));
			}
			if ( date >= 0 ) {
				config.addChild( "date", String.valueOf( date ));
			}
			if ( hour >= 0 ) {
				config.addChild( "hour", String.valueOf( hour ));
			}
			if ( min >= 0 ) {
				config.addChild( "min", String.valueOf( min ));
			}
			if ( sec >= 0 ) {
				config.addChild( "sec", String.valueOf( sec ));
			}
		}
	}
}

