package	jp.co.sds_corp.util.daemon.manager;

import java.io.FileOutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Properties;
import java.util.Vector;

import jp.co.sds_corp.util.common.CmnLog;
import jp.co.sds_corp.util.config.DOMConfiguration;
import jp.co.sds_corp.util.config.DOMWriter;
import jp.co.sds_corp.util.config.IConfiguration;
import jp.co.sds_corp.util.daemon.IInitializer;
import jp.co.sds_corp.util.daemon.IService;
import jp.co.sds_corp.util.daemon.config.ConfigUtility;
import jp.co.sds_corp.util.daemon.core.CronDaemon;
import jp.co.sds_corp.util.daemon.core.Daemon;
import jp.co.sds_corp.util.daemon.core.IDaemon;
import jp.co.sds_corp.util.daemon.core.ITerminatable;
import jp.co.sds_corp.util.daemon.core.TimerDaemon;
import jp.co.sds_corp.util.daemon.remote.IRemoteServer;
import jp.co.sds_corp.util.daemon.remote.RemoteFactory;

/**
* デーモンマネージャです。
* 指定されたXMLファイルから、デーモンの情報を取得してインスタンスを生成します。
* また、RMIのI/Fを持ち、遠隔オペレーションやデーモンの状態表示等を
* サポートします。
<P>
XMLファイルは以下のような形式です。
<PRE>
&lt;?xml version="1.0" encoding="UTF-8"?&gt;
&lt;Domain&gt;
</PRE>

Remote要素には、リモートI/Fの情報を記述します。
設定内容については、それぞれのリモートI/Fを参照してください。
{@link jp.co.sds_corp.util.daemon.remote.rmi.RMIServer},
{@link jp.co.sds_corp.util.daemon.remote.socket.SocketServer},
{@link jp.co.sds_corp.util.daemon.remote.jms.JMSServer}
<PRE>
&lt;Remote type="Socket"&gt;
	&lt;host&gt;localhost&lt;/host&gt;
	&lt;port&gt;1088&lt;/port&gt;
&lt;/Remote&gt;
</PRE>
旧バージョンとの互換性のため、Registry要素も使用可能になっています。Regisrty要素を指定した場合には、RMI I/Fが選択されます。
<PRE>
&lt;Remote&gt;
	&lt;host&gt;localhost&lt;/host&gt;
	&lt;port&gt;1088&lt;/port&gt;
	&lt;name&gt;DaemonManager&lt;/name&gt;
	&lt;CreateRegistry&gt;true&lt;/CreateRegistry&gt;
&lt;/Remote&gt;
</PRE>

&lt;Initialize&gt;要素には、起動時に（デーモンの起動前に）呼び出す
{@link IInitializer}の情報を記述します。(Option)
<PRE>
&lt;Initialize name="Log4jConifgurator"
	class="jp.co.sds_corp.util.daemon.manager.Log4jConfigurator"&gt;
	&lt;config&gt;properties/util/CmnLog.properties&lt;/config&gt;
&lt;/Initialize&gt;
</PRE>

以下が、デーモンの指定です。
<BR>
&lt;Daemon&gt;要素の指定方法については、デーモンのAPI仕様を参照してください。
{@link Daemon}, {@link CronDaemon}, {@link TimerDaemon}
<PRE>
&lt;Daemon name="Test" class="jp.co.sds_corp.util.daemon.core.TimerDaemon"&gt;
	&lt;Service name="Test" class="SERVICE"&gt;
		...
	&lt;/Service&gt;
&lt;/Daemon&gt;
&lt;/Domain&gt;
</PRE>
*/
public class DaemonManager
{
	/** RMI Registry を指定するXML要素名 */
	protected String KEY_REGISTRY = "Registry";
	/** Remote I/Fを指定するXML要素名 */
	protected String KEY_REMOTE = "Remote";
	/** 初期処理 を指定するXML要素名 */
	protected String KEY_INITIALIZER = "Initialize";
	/** Daemon を指定するXML要素名 */
	protected String KEY_DAEMON = "Daemon";

	protected LinkedHashMap<String,IDaemon>	daemonMap =
		new LinkedHashMap<String,IDaemon>();
	private IRemoteServer server;
	private boolean	bExit = false;
	private String config = "daemon.xml";
	private IConfiguration topConfig;
	private IConfiguration regConfig;
	private	LinkedHashMap<String,IInitializer> initMap =
		new LinkedHashMap<String,IInitializer>();
	private Properties	prop = new Properties();
	protected static DaemonManager instance;

	/**
	* デーモンマネージャを生成します。
	*/
	protected DaemonManager() {
		String app = System.getProperty( "app.name" );
		setProperty( "app.name", app );
		String msg = "Create. " + getClass().getName();
		if ( app != null ) {
			msg += " [" + app + "]";
		}
		CmnLog.debug( this, msg );
	}

	/**
	* デーモンマネージャのインスタンスを返します。
	*/
	public static DaemonManager getInstance() {
		if ( instance == null ) {
			instance = new DaemonManager();
			Thread shut = new Thread() {
				public void run() {
					instance.onShutdown();
				}
			};
			Runtime.getRuntime().addShutdownHook( shut );
		}
		return instance;
	}

	/**
	* デーモンマネージャを、指定されたXMLファイルで初期化します。
	*/
	public void init() throws Exception {
		if ( System.getProperty( "DaemonManager.config" ) != null ) {
			config = System.getProperty( "DaemonManager.config" );
		}
		init( new DOMConfiguration( config ));
	}

	/**
	* デーモンマネージャを、指定されたXMLファイルで初期化します。
	*/
	public void init( IConfiguration config ) throws Exception {
		topConfig = config;
		IConfiguration c;

		regConfig = config.getChild( KEY_REMOTE );
		if ( regConfig == null ) {
			// @@@ 互換性のため
			regConfig = config.getChild( KEY_REGISTRY );
			if ( regConfig != null ) {
				regConfig.setAttribute( "type", "RMI" );
			}
		}

		Iterator<IConfiguration> itr;
		itr = config.getChilds( KEY_INITIALIZER );
		initMap.putAll( createAllInitializer( itr ) );

		Iterator<String> itr2;
		itr2 = initMap.keySet().iterator();
		while ( itr2.hasNext()) {
			IInitializer init = (IInitializer )initMap.get( itr2.next());
			init.start();
		}

		itr = config.getChilds( KEY_DAEMON );
		daemonMap.clear();
		daemonMap.putAll( createAllDaemon( itr ));
		// addTerminatable( daemonMap );

		initRemote( regConfig );

		String msg = "Initialized. " + getClass().getName();
		String app = getProperty( "app.name" );
		if ( app != null ) {
			msg += " [" + app + "]";
		}
		CmnLog.info( this, msg );
	}

	/**
	* デーモンマネージャの{@link IRemoteServer}を生成します。
	* @param	config	設定情報
	*/
	protected void initRemote( IConfiguration config ) throws Exception {
		if ( config == null ) {
			return;
		}
		server = RemoteFactory.createServer( this, config );
	}

	/**
	* 指定された初期化処理({@link IInitializer})を全て生成して呼び出します。
	*/
	protected Map<String,IInitializer> createAllInitializer( Iterator<IConfiguration> itr )
		throws Exception
	{
		LinkedHashMap<String,IInitializer> ret = new LinkedHashMap<String,IInitializer>();
		while ( itr.hasNext()) {
			IConfiguration c = (IConfiguration )itr.next();
			IInitializer init = createInitializer( c );
			ret.put( c.getAttributes().get( "name" ), init );
		}
		return ret;
	}

	/**
	* 指定された初期化処理({@link IInitializer})を生成して呼び出します。
	*/
	protected IInitializer createInitializer( IConfiguration config )
		throws Exception
	{
		Object iObj = ConfigUtility.createInstance( config );
		IInitializer init = (IInitializer)iObj;
		CmnLog.debug( this, "loading Initializer: " + init.getClass().getName() );
		try {
			init.init( config );
		} catch ( Exception ex ) {
			CmnLog.error( this, "Initializer [" + init + "] init error. " );
			throw( ex );
		}
		return init;
	}

	/**
	* 指定されたデーモンを全て生成します。
	*/
	protected Map<String,IDaemon> createAllDaemon( Iterator<IConfiguration> itr )
		throws Exception
	{
		LinkedHashMap<String,IDaemon> ret = new LinkedHashMap<String,IDaemon>();

		while ( itr.hasNext()) {
			IConfiguration c = (IConfiguration )itr.next();
			int nDup = c.getInteger( "duplicate", 1 );
			for ( int i = 0; i < nDup; i++ ) {
				IDaemon dmn = createDaemon( c, i );
				addDaemon( ret, dmn );
			}
		}
		return ret;
	}

	/**
	* 指定されたデーモンを生成します。
	*/
	public IDaemon createDaemon( IConfiguration config, int i )
		throws Exception
	{
		int nDup = config.getInteger( "duplicate", 1 );
		IDaemon dmn = (IDaemon )ConfigUtility.createInstance( config );
		CmnLog.debug( this, "loading Daemon: " + dmn.getClass().getName() );
		String name;
		if ( nDup == 1 ) {
			name = config.getAttributes().get( "name" );
		} else {
			name = config.getAttributes().get( "name" ) + i;
		}
		dmn.setName( name );
		try {
			dmn.init( config );
		} catch ( Exception ex ) {
			CmnLog.error( this, "Daemon [" + dmn + "] init error. " );
			throw( ex );
		}
		return dmn;
	}

	/**
	* IInitializer名の一覧を取得します。
	* @return	IInitializerの一覧
	*/
	public Map<String,IInitializer> getInitializerMap() {
		return initMap;
	}

	/**
	* IInitializerを取得します。
	* @param	name		IInitializerの名前
	* @return	IInitializer
	*/
	public IInitializer getInitializer( String name ) {
		return (IInitializer )initMap.get( name );
	}

	/**
	* デーモン名の一覧を取得します。
	* @return	デーモンの一覧
	*/
	public Map<String,IDaemon> getDaemonMap() {
		return daemonMap;
	}

	/**
	* デーモンを取得します。
	* @param	name		デーモンの名前
	* @return	デーモン
	*/
	public IDaemon getDaemon( String name ) {
		return (IDaemon )daemonMap.get( name );
	}

	/**
	* デーモンを追加します。
	* @param	dmn			デーモン
	*/
	public void addDaemon( Map<String,IDaemon> map, IDaemon dmn ) throws Exception {
		String name = dmn.getName();
		if ( map.get( name ) != null ) {
			throw( new Exception( "key duplicated [" + name + "]" ));
		}
		map.put( name, dmn );
		addTerminatable( dmn );
	}

	/**
	* デーモンを削除します。
	* @param	name		デーモンの名前
	*/
	public IDaemon removeDaemon( String name ) {
		IDaemon dmn = (IDaemon )daemonMap.remove( name );
		return dmn;
	}

	/**
	* 設定情報を取得します。
	* @return	設定情報
	*/
	public IConfiguration getConfig() {
		return topConfig;
	}

	/**
	* 全てのデーモンの情報をファイルにsaveします。
	*/
	public void save( String file, String name ) throws Exception {
		DOMConfiguration config = DOMConfiguration.createDocument( name );

		saveConfig( config );

		Writer out = new OutputStreamWriter(
			new FileOutputStream( file ), "UTF-8" );
		DOMWriter writer = new DOMWriter( out );
		writer.write( config, name );
		out.flush();
		out.close();
	}

	/**
	* 全てのデーモンの情報を格納します。
	*/
	public void saveConfig( IConfiguration config ) {
		copyChilds( config, "Remote" );
		copyChilds( config, "HtmlAdaptor" );
		copyChilds( config, "MBean" );
		copyChilds( config, "Initialize" );

		Iterator<String> itr = daemonMap.keySet().iterator();
		while ( itr.hasNext()) {
			IDaemon dmn = (IDaemon )daemonMap.get( itr.next() );
			IConfiguration c = config.addChild( "Daemon" );
			dmn.saveConfig( c );
		}
	}

	protected void copyChilds( IConfiguration config, String key ) {
		Iterator<IConfiguration> itr;
		itr = topConfig.getChilds( key );
		while ( itr.hasNext()) {
			IConfiguration c = (IConfiguration )itr.next();
			config.addChild( c );
		}
	}

	/**
	* 全てのデーモンを開始します。
	*/
	public void start() {
		Iterator<String> itr = daemonMap.keySet().iterator();
		while ( itr.hasNext()) {
			IDaemon dmn = (IDaemon )daemonMap.get( itr.next() );
			try {
				if ( !dmn.isAlive() ) {
					CmnLog.debug( this, "start: " + dmn.getName() + ":"
									+ dmn.getClass().getName() );
					dmn.start();
				}
			} catch ( Throwable ex ) {
				CmnLog.error( this, "Daemon module [" + dmn + "] cannot start. ", ex );
			}
		}
		// addTerminatable( daemonMap );

		String msg = "Started. " + getClass().getName();
		String app = getProperty( "app.name" );
		if ( app != null ) {
			msg += " [" + app + "]";
		}
		CmnLog.info( this, msg );
	}

	/**
	* 全てのデーモンを停止します。
	*/
	public void stop() {
		CmnLog.debug( this, "stopping " + getClass().getName() );
		ITerminatable t;
		Iterator<String> itr;

		for ( int i = 0; i < terminatables.size(); i++ ) {
			t = (ITerminatable )terminatables.get( i );
			if ( t instanceof IDaemon ) {
				CmnLog.debug( this, "stopping " + ((IDaemon)t).getName() );
			} else {
				CmnLog.debug( this, "stopping " + t );
			}
			try {
				t.terminate();
			} catch ( Throwable ex ) {
				CmnLog.error( this, "Daemon module [" + t + "] terminate error. ", ex );
			}
		}
		for ( int i = 0; i < terminatables.size(); i++ ) {
			t = (ITerminatable )terminatables.get( i );
			try {
				t.joinAll();
			} catch ( Throwable ex ) {
				CmnLog.error( this, "Daemon module [" + t + "] joinAll error. ", ex );
			}
			if ( t instanceof IDaemon ) {
				CmnLog.debug( this, "stopped " + ((IDaemon)t).getName() );
			} else {
				CmnLog.debug( this, "stopped " + t );
			}
		}
		terminatables.clear();

		itr = initMap.keySet().iterator();
		while ( itr.hasNext()) {
			IInitializer init = (IInitializer )initMap.get( itr.next());
			init.stop();
		}

		if ( server != null ) {
			server.exit();
		}

		String msg = "stopped " + getClass().getName();
		String app = getProperty( "app.name" );
		if ( app != null ) {
			msg += " [" + app + "]";
		}
		CmnLog.error( this, msg );
		setExit( true );
	}

	/**
	* 全てのデーモンを停止して、デーモンマネージャーを終了します。
	*/
	public void quit() {
		stop();
	}

	private Vector<ITerminatable> terminatables = new Vector<ITerminatable>();

	/** 終了時にフックしてほしいモジュールを登録します。 */
	public void addTerminatable( ITerminatable t ) {
		terminatables.add( t );
	}

	/**
	* 終了時にフックしてほしいモジュールを登録します。
	* List中で{@link ITerminatable}を実装していないオブジェクトは
	* スキップされます。
	* @param	map		フックするモジュールを格納したMap
	*/
	public void addTerminatable( Map<String,ITerminatable> map ) {
		Iterator<String> itr = map.keySet().iterator();
		while ( itr.hasNext()) {
			ITerminatable term = map.get( itr.next() );
			terminatables.add( term );
		}
	}

	/**
	* 終了時の処理を行ないます。
	*/
	public void onShutdown() {
		if ( !isExit() ) {
			stop();
		}
	}

	protected synchronized void setExit( boolean var ) {
		bExit = var;
	}
	protected synchronized boolean isExit() {
		return bExit;
	}

	/**
	* 指定されたキーを持つプロパティの値を設定します。
	* @param	key		キー
	* @param	value	プロパティの値
	*/
	public String setProperty( String key, String value ) {
		return (String )prop.setProperty( key, value );
	}
	/**
	* 指定されたキーを持つプロパティの値を取得します。
	* @param	key		キー
	* @return	プロパティの値
	*/
	public String getProperty( String key ) {
		return prop.getProperty( key );
	}
	/**
	* 指定されたキーを持つプロパティの値を取得します。
	* @param	key		キー
	* @param	defval	デフォルト値
	* @return	プロパティの値
	*/
	public String getProperty( String key, String defval ) {
		return prop.getProperty( key, defval );
	}
	/**
	* プロパティを取得します。
	* @return	プロパティ
	*/
	public Properties getProperty() {
		return prop;
	}

	/**
	* デーモン起動中のmain threadです。
	* 実際には何もしていません。
	*/
	public void mainLoop() {
		try {
			Thread.currentThread().setName( "DaemonManager main" );
			for ( ;; ) {
				Thread.sleep( 1000 );
				if ( isExit() ) {
					break;
				}
				Iterator<String> itr = daemonMap.keySet().iterator();
				while ( itr.hasNext()) {
					IDaemon dmn = (IDaemon )daemonMap.get( itr.next() );
					IService svc = dmn.getService();
					if ( svc != null ) {
						svc.setValid( dmn.isValid() );
					}
				}
				Thread.yield();
			}
		} catch ( Exception ex ) {
			ex.printStackTrace();
			CmnLog.error( this, "mainLoop error", ex );
		}
	}

	/**
	* デーモンマネージャの起動処理です。
	* @param	args	コマンドラインパラメータ
	*/
	public static void main( String[] args ) {
		DaemonManager obj = null;
		try {
			obj = DaemonManager.getInstance();

			obj.init();
			obj.start();
			obj.mainLoop();
			System.exit( 0 );
		} catch ( Throwable ex ) {
			CmnLog.error( obj, "error", ex );
			ex.printStackTrace();
			System.exit( 9 );
		}
	}
}
