package	jp.co.sds_corp.util.daemon.remote.rmi;

import java.lang.reflect.Method;
import java.net.MalformedURLException;
import java.rmi.AlreadyBoundException;
import java.rmi.RemoteException;
import java.util.Map;

import jp.co.sds_corp.util.common.CmnLog;
import jp.co.sds_corp.util.config.IConfiguration;
import jp.co.sds_corp.util.daemon.IService;
import jp.co.sds_corp.util.daemon.core.IDaemon;
import jp.co.sds_corp.util.daemon.core.Interval;
import jp.co.sds_corp.util.daemon.core.Schedule;
import jp.co.sds_corp.util.remote.AuthKey;
import jp.co.sds_corp.util.remote.AuthUtility;
import jp.co.sds_corp.util.remote.ProcessRemoteControlImpl;
import jp.co.sds_corp.util.remote.ServerProcess;

/**
* デーモンマネージャのRMI I/Fの実装です。
*/
public class DaemonRemoteImpl extends ProcessRemoteControlImpl
	implements DaemonRemote
{
	/**
	* 指定された情報で初期化します。
	* @param	server	デーモンマネージャのサーバ実装の参照
	* @param	host	RMI Registry のホスト名
	* @param	port	RMI Registry のポート番号
	* @param	name	RMI Registry への登録名
	* @param	bCr		RMI Registry をcreateするかどうかのフラグ
	*/
	public DaemonRemoteImpl( ServerProcess server,
			String host, int port, String name, boolean bCr )
		throws RemoteException, MalformedURLException, AlreadyBoundException
	{
		super( server, host, port, name, bCr );
	}

	/**
	* リモートコマンドを実行します。
	* @param	key		RMI 認証キー
	* @param	cmd		実行するコマンド。（メソッド名）
	* @param	cls		パラメータのクラス
	* @param	prm		パラメータ
	*/
	public Object command( AuthKey key, String cmd, Class<?>[] cls, Object[] prm )
		throws Exception
	{
		AuthUtility.isValidSessionID( key );

		if ( isDebug()) {
			String msg = "command:" + cmd + "(";
			if ( cls != null ) {
				for ( int i = 0; i < cls.length; i++ ) {
					if ( i > 0 ) {
						msg += ",";
					}
					if ( prm.length >= (i+1) ) {
						msg += prm[i];
					}
					msg += "[" + cls[i].getName() + "]";
				}
			}
			msg += ")";
			CmnLog.debug( this, msg );
		}
		ServerProcess server = getServer();
		Method mtd = server.getClass().getMethod( cmd, cls );
		Object ret = mtd.invoke( server, prm );
		return ret;
	}

	/**
	* リモートコマンドを実行します。
	* @param	key		RMI 認証キー
	* @param	cmd		実行するコマンド。（メソッド名）
	*/
	public Object command( AuthKey key, String cmd ) throws Exception {
		Class<?>[] cls = {};
		Object[] prm = {};
		Object ret = command( key, cmd, cls, prm );
		return ret;
	}

	/**
	* リモートコマンドを実行します。
	* @param	key		RMI 認証キー
	* @param	cmd		実行するコマンド。（メソッド名）
	* @param	prm1	パラメータ
	*/
	public Object command( AuthKey key, String cmd, Object prm1 )
		throws Exception
	{
		Class<?>[] cls = { prm1.getClass() };
		Object[] prm = { prm1 };
		Object ret = command( key, cmd, cls, prm );
		return ret;
	}

	/**
	* リモートコマンドを実行します。
	* @param	key		RMI 認証キー
	* @param	cmd		実行するコマンド。（メソッド名）
	* @param	prm1	パラメータ
	* @param	prm2	パラメータ
	*/
	public Object command( AuthKey key, String cmd, Object prm1, Object prm2 )
		throws Exception
	{
		Class<?>[] cls = { prm1.getClass(), prm2.getClass() };
		Object[] prm = { prm1, prm2 };
		Object ret = command( key, cmd, cls, prm );
		return ret;
	}

	/**
	* 全てのデーモンを開始します。
	* @param	key		RMI 認証キー
	*/
	public void start( AuthKey key ) throws Exception {
		command( key, "start" );
	}

	/**
	* 指定された名前のデーモンを開始します。
	* @param	key		RMI 認証キー
	* @param	name	デーモンの名前
	*/
	public void start( AuthKey key, String name ) throws Exception {
		command( key, "start", name );
	}

	/**
	* 全てのデーモンを停止します。
	* @param	key		RMI 認証キー
	*/
	public void stop( AuthKey key ) throws Exception {
		command( key, "stop" );
	}

	/**
	* 指定された名前のデーモンを停止します。
	* @param	key		RMI 認証キー
	* @param	name	デーモンの名前
	*/
	public void stop( AuthKey key, String name ) throws Exception {
		command( key, "stop", name );
	}

	/**
	* 全てのデーモンを停止して、デーモンマネージャを終了します。
	* @param	key		RMI 認証キー
	*/
	public void quit( AuthKey key ) throws Exception {
		command( key, "quit" );
	}

	/**
	* 全てのデーモンの情報を再読み込みします。
	* @param	key		RMI 認証キー
	*/
	public void reload( AuthKey key ) throws Exception {
		command( key, "reload" );
	}

	/**
	* 全てのデーモンを初期化します。
	* @param	key		RMI 認証キー
	*/
	public void initialize( AuthKey key ) throws Exception {
		command( key, "initialize" );
	}

	/**
	* 指定された名前のデーモンを初期化します。
	* @param	key		RMI 認証キー
	* @param	name	デーモンの名前
	*/
	public void initialize( AuthKey key, String name ) throws Exception {
		command( key, "initialize", name );
	}

	/**
	* 全てのデーモンの情報を文字列で取得します。
	* @param	key		RMI 認証キー
	*/
	public Object status( AuthKey key ) throws Exception {
		return command( key, "status" );
	}

	/**
	* 指定された名前のデーモンの情報を文字列で取得します。
	* @param	key		RMI 認証キー
	* @param	name	デーモンの名前
	*/
	public Object status( AuthKey key, String name ) throws Exception {
		return command( key, "status", name );
	}

	/**
	* 全ての Thread の情報を文字列で取得します。
	* @param	key		RMI 認証キー
	*/
	public Object thread( AuthKey key ) throws Exception {
		return command( key, "thread" );
	}
	/**
	* JVMの情報を文字列で取得します。
	* @param	key		RMI 認証キー
	*/
	public Object vminfo( AuthKey key ) throws Exception {
		return command( key, "vminfo" );
	}
	/**
	* JVMの情報を文字列で取得します。
	* @param	key		RMI 認証キー
	*/
	public Object properties( AuthKey key ) throws Exception {
		return command( key, "properties" );
	}
	/**
	* Garbage Collectionを実施します。
	* @param	key		RMI 認証キー
	*/
	public void gc( AuthKey key ) throws Exception {
		command( key, "gc" );
	}
	/**
	* 設定情報を出力します。
	* @param	key		RMI 認証キー
	* @param	file	ファイル名
	* @param	name	root node名
	*/
	public void save( AuthKey key, String file, String name )
		throws Exception
	{
		Class<?>[] cls = { String.class, String.class };
		Object[] prm = { file, name };
		command( key, "save", cls, prm );
	}

	/**
	* サービスの一覧を返します。
	* @param	key		RMI 認証キー
	* @return	サービスの一覧
	*/
	public Map getServiceMap( AuthKey key ) throws Exception {
		return (Map )command( key, "getServiceMap" );
	}
	/**
	* 指定された名前のサービスを返します。
	* @param	key		RMI 認証キー
	* @param	name	サービスの名前
	* @return	指定されたサービス
	*/
	public IService getService( AuthKey key, String name ) throws Exception {
		return (IService )command( key, "getService", name );
	}
	/**
	* 指定された名前のサービスの状態を返します。
	* @param	key		RMI 認証キー
	* @param	name	サービスの名前
	* @return	指定されたサービス
	*/
	public IService.Status getServiceStatus( AuthKey key, String name )
		throws Exception
	{
		IService svc = (IService )command( key, "getService", name );
		if ( svc == null ) {
			return IService.Status.UNKNOWN;
		}
		return svc.getStatus();
	}

	/**
	* デーモンの一覧を返します。
	* @param	key		RMI 認証キー
	* @return	デーモンの一覧
	*/
	public Map getDaemonMap( AuthKey key ) throws Exception {
		return (Map )command( key, "getDaemonMap" );
	}
	/**
	* 指定された名前のデーモンを返します。
	* @param	key		RMI 認証キー
	* @param	name	デーモンの名前
	* @return	指定されたデーモン
	*/
	public IDaemon getDaemon( AuthKey key, String name ) throws Exception {
		return (IDaemon )command( key, "getDaemon", name );
	}
	/**
	* 指定された名前のデーモンの状態を返します。
	* @param	key		RMI 認証キー
	* @param	name	デーモンの名前
	* @return	指定されたデーモン
	*/
	public IDaemon.Status getDaemonStatus( AuthKey key, String name )
		throws Exception
	{
		IDaemon dmn = (IDaemon )command( key, "getDaemonStatus", name );
		if ( dmn == null ) {
			return IDaemon.Status.UNKNOWN;
		}
		return dmn.getStatus();
	}

	/**
	* 全てのデーモンの情報を文字列で取得します。
	* @param	key		RMI 認証キー
	*/
	public IDaemon addDaemon( AuthKey key, IConfiguration config )
		throws Exception
	{
		Class<?>[] cls = { IConfiguration.class };
		Object[] prm = { config };
		Object ret = command( key, "addDaemon", cls, prm );
		return (IDaemon )ret;
	}

	/**
	* 指定された名前のデーモンを返します。
	* @param	key		RMI 認証キー
	* @param	name	デーモンの名前
	* @return	指定されたデーモン
	*/
	public Object invokeDaemon( AuthKey key, String name, String method )
		throws Exception
	{
		return command( key, "invodeDaemon", name, method );
	}

	/**
	* CronDaemonにスケジュールを追加します。
	* @param	key		RMI 認証キー
	* @param	name	デーモンの名前
	* @param	sch		スケジュール
	*/
	public void addSchedule( AuthKey key, String name, Schedule sch )
		throws Exception
	{
		command( key, "addSchedule", name, sch );
	}

	/**
	* CronDaemonのスケジュールを削除します。
	* @param	key		RMI 認証キー
	* @param	name	デーモンの名前
	* @param	idx		削除するスケジュールのインデックス
	* @return	削除されたスケジュール
	*/
	public Schedule removeSchedule( AuthKey key, String name, Integer idx )
		throws Exception
	{
		return (Schedule )command( key, "removeSchedule", name, idx );
	}

	/**
	* TimerDaemonに間隔を追加します。
	* @param	key		RMI 認証キー
	* @param	name	デーモンの名前
	* @param	itv		間隔
	*/
	public void addInterval( AuthKey key, String name, Interval itv )
		throws Exception
	{
		command( key, "addInterval", name, itv );
	}

	/**
	* TimerDaemonの間隔を削除します。
	* @param	key		RMI 認証キー
	* @param	name	デーモンの名前
	* @param	idx		削除するスケジュールのインデックス
	* @return	削除されたスケジュール
	*/
	public Interval removeInterval( AuthKey key, String name, Integer idx )
		throws Exception
	{
		return (Interval )command( key, "removeInterval", name, idx );
	}

	/**
	* CronDaemon,TimerDaemonの処理を直ちに実行します。
	* @param	key		RMI 認証キー
	* @param	name	デーモンの名前
	*/
	public void startImmediate( AuthKey key, String name )
		throws Exception
	{
		command( key, "startImmediate", name );
	}

	/**
	* helpを表示します。
	* @param	key		RMI 認証キー
	*/
	public Object help( AuthKey key ) throws Exception {
		return command( key, "help" );
	}

	boolean	debug;	// debugフラグ

	/**
	* debugフラグを設定します。
	* @param	var		設定するデータ
	*/
	public void setDebug( boolean var ) {
		this.debug = var;
	}
	/**
	* debugフラグを取得します。
	* @return	debugフラグ
	*/
	public boolean isDebug() {
		return debug;
	}

}
