package	jp.co.sds_corp.util.remote;

import java.lang.reflect.Method;
import java.rmi.Naming;

/**
* プロセスコントローラのクライアント機能の実装を提供します。
* 実際にはこのクラスでは個別のコマンドには対応していませんので
* このクラスを継承したクラスで、個別のコマンドに対応する必要があります。
*/
public class ProcessController
{
	String	user = null;
	String	passwd = null;
	String	host = null;
	int		port;
	String	name;
	ProcessRemoteControl remote;
	AuthKey key;
	boolean	connected = false;	// 接続しているか

	/**
	* RMIクライアント機能を初期化します。
	*/
	public ProcessController() {
	}

	/**
	* RMIクライアント機能を初期化します。
	* @param	host	RMI Registry のホスト名
	* @param	port	RMI Registry のポート番号
	* @param	user	RMI サーバのユーザID
	* @param	passwd	RMI サーバのパスワード
	*/
	public ProcessController( String host, int port, String name,
				String user, String passwd )
		throws Exception
	{
		this.host = host;
		this.port = port;
		this.name = name;
		this.user = user;
		this.passwd = passwd;
	}

	/**
	* 設定された情報でRMIクライアント機能を初期化します。
	*/
	public void connect() throws Exception {
		remote = (ProcessRemoteControl)
			Naming.lookup( "rmi://" + host + ":" + port + "/" + name );

		key = new AuthKey( user, passwd );
		key = remote.auth( key );
		connected = true;
	}

	/**
	* 設定された情報でRMIクライアント機能を終了します。
	*/
	public void disconnect() throws Exception {
		remote = null;
		connected = false;
	}

	/**
	* 接続しているかを設定します。
	* @param	var		設定するデータ
	*/
	public void setConnected( boolean var ) {
		this.connected = var;
	}
	/**
	* 接続しているかを取得します。
	* @return	接続しているか
	*/
	public boolean isConnected() {
		return connected;
	}


	/**
	* Registryホスト(host)を設定します。
	* @param	var		設定するデータ
	*/
	public void setHost( String var ) {
		this.host = var ;
	}
	/**
	* Registryホスト(host)を取得します。
	* @return	hostの内容
	*/
	public String getHost() {
		return host;
	}

	/**
	* Registryポート(port)を設定します。
	* @param	var		設定するデータ
	*/
	public void setPort( int var ) {
		this.port = var ;
	}
	/**
	* Registryポート(port)を取得します。
	* @return	portの内容
	*/
	public int getPort() {
		return port;
	}

	/**
	* Registry登録名(name)を設定します。
	* @param	var		設定するデータ
	*/
	public void setName( String var ) {
		this.name = var ;
	}
	/**
	* Registry登録名(name)を取得します。
	* @return	nameの内容
	*/
	public String getName() {
		return name;
	}

	/**
	* ユーザID(user)を設定します。
	* @param	var		設定するデータ
	*/
	public void setUser( String var ) {
		this.user = var ;
	}
	/**
	* ユーザID(user)を取得します。
	* @return	userの内容
	*/
	public String getUser() {
		return user;
	}

	/**
	* パスワード(passwd)を設定します。
	* @param	var		設定するデータ
	*/
	public void setPasswd( String var ) {
		this.passwd = var ;
	}
	/**
	* パスワード(passwd)を取得します。
	* @return	passwdの内容
	*/
	public String getPasswd() {
		return passwd;
	}

	/**
	* リモートのメソッドを呼び出します。
	* @param	cmd		メソッド名
	* @param	cls		パラメータのクラス
	* @param	prm		パラメータのオブジェクト
	*/
	public Object execRemote( String cmd, Class<?>[] cls, Object[] prm )
		throws Exception
	{
/*
		if ( !connected ) {
			init();
		}
*/
		Class<?>[] c = new Class[ cls.length + 1 ];
		Object[] p = new Object[ cls.length + 1 ];
		c[0] = AuthKey.class;
		p[0] = key;
		for ( int i = 0; i < cls.length; i++ ) {
			c[i+1] = cls[i];
			p[i+1] = prm[i];
		}

		Method mtd = remote.getClass().getMethod( cmd, c );
		Object obj = mtd.invoke( remote, p );
		return obj;
	}

	/**
	* リモートのメソッドを呼び出します。
	* @param	cmd		メソッド名
	* @param	prm1	パラメータ
	*/
	public Object execRemote( String cmd, String prm1 )
		throws Exception
	{
		Class<?>[] cls = { String.class };
		Object[] prm = { prm1 };
		return execRemote( cmd, cls, prm );
	}

	/**
	* リモートのメソッドを呼び出します。
	* @param	cmd		メソッド名
	*/
	public Object execRemote( String cmd )
		throws Exception
	{
		Class<?>[] cls = {};
		Object[] prm = {};
		return execRemote( cmd, cls, prm );
	}

	/**
	* コマンドラインからパラメータを読み込み、RMIクライアントを起動します。
	* @param	args	コマンドラインパラメータ
	*/
	public static void main( String[] args ) {
		if ( args.length < 3 ) {
			usage();
		}
		try {
			ProcessController ctl = new ProcessController(
				args[0], Integer.valueOf( args[1] ).intValue(), args[2],
				null, null );
			// ctl.setUser( user );
			// ctl.setPasswd( passwd );
			ctl.connect();

			Object ret = null;
			switch ( args.length ) {
			case 4:
				ret = ctl.execRemote( args[3] );
				break;
			default:
				String cmd = args[3];
				Class<?>[] cls = new Class[args.length-4];
				Object[] prm = new Object[args.length-4];
				for ( int i = 4; i < args.length; i++ ) {
					cls[i-4] = args[i].getClass();
					prm[i-4] = args[i];
				}
				ret = ctl.execRemote( cmd, cls, prm );
				break;
			}
			if ( ret != null ) {
				System.out.println( ret );
			}
			System.exit( 0 );
		} catch ( Throwable ex ) {
			// fixme
			ex.printStackTrace();
			System.exit(1);
		}
	}

	private static void usage() {
		System.out.println(
			"usage: java ProcessController " +
			"host port name action [param]");
			System.exit(0);
	}
}
